"""plugins.py contains the main type and base class used by the analyzis plugins.

It also contains the work functions used to load the plugins both from disc and
from the resources."""

from act.scio import plugins
import addict
from importlib import import_module
from importlib.machinery import ModuleSpec
from importlib.util import module_from_spec, spec_from_file_location
from pydantic import BaseModel, StrictStr
from typing import Text, List, Optional
import logging
import os
import pkgutil


module_interface = ["name", "analyze", "info", "version", "dependencies"]


class Result(BaseModel):
    """The result type returned by all analyze methods of the plugins."""

    name: StrictStr
    version: StrictStr
    result: addict.Dict


class BasePlugin:
    """The class that all analyzis plugins inherits. Contains the basis attributes and
    interface required by the plugin system."""

    def __init__(self: object):
        pass

    name = "BasePlugin"
    info = "This is the empty plugin of a plugin for Scio"
    version = "0.1"
    dependencies: List[Text] = []
    configdir = ""
    debug = False

    async def analyze(self, nlpdata: addict.Dict) -> Result:
        """Main analyzis method"""
        return Result(name=self.name, version=self.version, result=addict.Dict({"test": nlpdata.content}))


def load_default_plugins() -> List[BasePlugin]:
    """load_default_plugins scans the package for internal plugins, loading
    them dynamically and checking for the presence of the attributes defined in
    module_interface"""

    myplugins: List[BasePlugin] = []

    prefix = plugins.__name__ + "."
    for _, modname, _ in pkgutil.iter_modules(plugins.__path__, prefix):

        p = load_plugin(modname)
        if p:
            myplugins.append(p)

    return myplugins


def load_external_plugins(directory: Text) -> List[BasePlugin]:
    """load_external_plugins scans a directory for .py files, and attempts to
    import each file, adding them to the list of modules. The functions will
    only add the module to the returned list of modules if it has a dictionary
    describing the module_interface list"""

    myplugins: List[BasePlugin] = []

    for plugin_file_name in os.listdir(directory):
        if plugin_file_name == "__init__.py":
            continue
        plugin_path = os.path.join(directory, plugin_file_name)
        if os.path.isfile(plugin_path) and plugin_path.endswith(".py"):

            p = load_plugin(plugin_path)
            if p:
                myplugins.append(p)

    return myplugins


def load_plugin(module_name: Text) -> Optional[BasePlugin]:
    if module_name.endswith(".py"):
        spec: ModuleSpec = spec_from_file_location("plugin_mod", module_name)
        module = module_from_spec(spec)
        spec.loader.exec_module(module)  # type: ignore
    else:
        try:
            module = import_module(module_name)
        except Exception as e:
            logging.warning(e)
            return None

    conform = True
    try:
        p: BasePlugin = module.Plugin()  # type: ignore
    except AttributeError as err:
        logging.warning("Could not load plugin from module %s: %s", module_name, err)
        return None

    for mint in module_interface:
        if not hasattr(p, mint):
            logging.warning("%s does not have %s attribute", p.name, mint)
            conform = False

    if not conform:
        return None

    return p



#!/usr/bin/env python3

""" SCIO Analyze module """

from act.scio import plugin
from typing import Optional, List

import addict  # type: ignore
import argparse
import asyncio
import greenstalk  # type: ignore
import gzip
import json
import logging
import os.path
import requests
import sys

import caep

import act.scio.logsetup
import act.scio.config

def parse_args() -> argparse.Namespace:
    """Helper setting up the argsparse configuration"""

    arg_parser = act.scio.config.parse_args("Scio 2 Analyzer")
    arg_parser.add_argument('--plugins', dest='plugins', type=str)
    arg_parser.add_argument('--webdump', dest='webdump', type=str, help="URI to post result data")

    return caep.config.handle_args(arg_parser, "scio/etc", "scio.ini", "analyze")


def get_input(beanstalk_client: Optional[greenstalk.Client] = None) -> addict.Dict:
    """Helper function to abstract away how we get the text to work on"""

    nlpdata = addict.Dict()
    if not beanstalk_client:
        logging.info("Waiting for work on stdin")
        data = sys.stdin.read()
        nlpdata.content = data
    else:
        # ADD BEANSTALK JOB CONSUMPTION
        logging.info("Waiting for work from beanstalk")
        job = beanstalk_client.reserve()
        try:
            nlpdata = addict.Dict(json.loads(gzip.decompress(job.body)))
            logging.info(nlpdata.keys())
        except OSError as e:
            # File not found - log error
            logging.error(e)
        finally:
            # Remove job, either on success or file not found
            beanstalk_client.delete(job)

    return nlpdata


async def analyze(plugins: List[plugin.BasePlugin],
                  beanstalk_client: Optional[greenstalk.Client] = None) -> addict.Dict:
    """Main analyze loop running all plugins on the text"""

    loop = asyncio.get_event_loop()

    nlpdata: addict.Dict = get_input(beanstalk_client)

    staged = []  # for plugins with dependencies
    pipeline = []  # for plugins to be run now

    for p in plugins:
        if p.dependencies:
            staged.append(p)
        else:
            pipeline.append(p)

    while pipeline:
        tasks = []
        for p in pipeline:
            tasks.append(loop.create_task(p.analyze(nlpdata)))

        await asyncio.gather(*tasks)

        for task in tasks:
            if task.exception():
                logging.warning("%s return an exception: %s", task, task.exception())
            else:
                res = task.result()
                nlpdata[res.name] = res.result

        pipeline = []
        for candidate in staged[:]:
            if all(dep in nlpdata for dep in candidate.dependencies):
                pipeline.append(candidate)
                staged.remove(candidate)

    for candidate in staged:
        logging.warning("Candidate %s did not run due to unmet dependency %s",
                        candidate.name,
                        candidate.dependencies)

    return nlpdata


async def async_main() -> None:
    """Async version of main"""

    args = parse_args()

    act.scio.logsetup.setup_logging(args.loglevel, args.logfile, "scio-analyze")

    plugins = plugin.load_default_plugins()

    if args.plugins:
        try:
            plugins += plugin.load_external_plugins(args.plugins)
        except FileNotFoundError:
            logging.warning("Unable to load plugins from %s", args.plugins)

    # Inject config directory into each plugin
    for p in plugins:
        p.configdir = os.path.join(args.config_dir, "etc/plugins")

    beanstalk_client = act.scio.config.beanstalk_client(args, watch="scio_analyze")
    elasticsearch_client = act.scio.config.elasticsearch_client(args)

    loop = asyncio.get_event_loop()

    while True:
        task = loop.create_task(analyze(plugins, beanstalk_client))
        try:
            await task
        except LookupError as e:
            logging.error("Got LookupError. If nltk data is missing, run scio-nltk-download, which should download all nltk data to ~/nltk_data.")
            raise

        result = task.result()
        result_json = json.dumps(result, indent="  ")

        if args.webdump:
            proxies = {
                'http': args.proxy_string,
                'https': args.proxy_string
            } if args.proxy_string else None

            r = requests.post(args.webdump, data=result_json, proxies=proxies)
            if r.status_code != 200:
                logging.error("Unable to post result data to webdump: %s", r.text)

        elif elasticsearch_client:
            hexdigest = result.get("hexdigest")

            if not hexdigest:
                logging.error("Missing hexdigest, skipping elasticsearch storage")
            else:
                elasticsearch_client.index(index="scio2", id=hexdigest, body=result)
                logging.info("Stored %s to elasticsearch", hexdigest)

        else:
            print(result_json)

        # If we are not listening on a beanstalk work queue, behave like a command line
        # utility and exit after one document.
        if not beanstalk_client:
            break


def main() -> None:
    """Main entry point"""

    loop = asyncio.get_event_loop()
    loop.run_until_complete(asyncio.wait([async_main()]))


if __name__ == "__main__":
    main()



"""Helper functions related to downloading feeds and files"""

from typing import Text, Optional, Any, Dict, cast, List, Tuple
import argparse
import concurrent.futures
import logging
import os.path
import requests
import shutil
import urllib.parse

import feedparser  # type: ignore

from act.scio.feeds import analyze, extract


def download_and_store(
        feed_url: Text,
        ignore_file: Optional[Text],
        storage_path: Text,
        link: urllib.parse.ParseResult) -> Text:
    """Download and store a link. Storage defined in args"""

    # check if the actual url is in the ignore file. If so, no download will take place.
    if analyze.in_ignore_file(link.geturl(), ignore_file):
        logging.info("Download link [%s] in ignore file.", link.geturl())
        return ""

    logging.info("downloading %s", link.geturl())

    # if netloc does not contain a hostname, assume a relative path to the feed url
    if link.netloc == '':
        parsed_feed_url = urllib.parse.urlparse(feed_url)
        link = link._replace(scheme=parsed_feed_url.scheme,
                             netloc=parsed_feed_url.netloc,
                             path=link.path)
        logging.info("possible relative path %s, trying to append host: %s",
                     link.path, parsed_feed_url.netloc)

    req = requests.get(link.geturl(),
                       headers=default_headers(),
                       verify=False,
                       stream=True,
                       timeout=60)

    if req.status_code >= 400:
        logging.info("Status %s - %s", req.status_code, link)
        return ""

    fname = os.path.join(storage_path,
                         "download",
                         extract.safe_filename(os.path.basename(link.path)))

    # check if the filename on disk is in the ignore file. If so, do not return filename
    # for upload. This differ from URL in the ignore file as the file is in fact downloaded by
    # the feed worker, but not uploaded to Scio.
    if analyze.in_ignore_file(fname, ignore_file):
        logging.info("Ignoring %s based on %s", fname, ignore_file)
        return ""

    with open(fname, "wb") as download_file:
        logging.info("Writing %s", fname)
        req.raw.decode_content = True
        shutil.copyfileobj(req.raw, download_file)

    return fname


def get_feed(feed_url: Text) -> Any:
    """Download and parse a feed"""

    feed_url = feed_url.strip()

    logging.info("Opening feed : %s", feed_url)

    req = requests.get(feed_url, headers=default_headers(), verify=False, timeout=60)

    return feedparser.parse(req.text)


def default_headers() -> Dict:
    """ Return default headers with a custom user agent """

    headers = requests.utils.default_headers()  # type: ignore
    headers.update({'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) '
                                  'AppleWebKit/537.36 (KHTML, like Gecko) '
                                  'Chrome/74.0.3729.169 Safari/537.36'})  # Chrome v74 2020-06-25

    return cast(Dict, headers)


def handle_feed(args: argparse.Namespace,
                feed_url: Text,
                partial: bool) -> Tuple[Text, Text, List[Text]]:
    """Take a feed, extract all entries, download the full original
    web page if partial , extract links and download any documents references
    if specified in the arguments and write the feed entry content
    to disk together with a meta data json file"""

    feed = get_feed(feed_url)

    if not feed:
        return "NOT FEED", feed_url, []

    files = []

    logging.info("%s contains %s entries",
                 feed_url,
                 len(feed["entries"]))

    for entry_n, entry in enumerate(feed["entries"]):
        logging.info("Handling : %s of %s : %s",
                     entry_n, len(feed["entries"]), entry['title'])

        filename, html_data = (extract.partial_entry_text_to_file(args, entry) if partial else
                               extract.entry_text_to_file(args, entry))
        files.append(filename)

        if not (filename and html_data):
            continue

        links = extract.get_links(entry, html_data)

        # Download all urls that looks like they have the correct file extension
        # and add the filenames of the downloaded files to the list of candidates
        # to upload.
        files.extend(download_and_store(feed_url, args.ignore, args.store_path, link)
                     for link in analyze.filter_links(args.file_format, links))

    return "OK", feed_url, list(filter(None, files))


def download_feed_list(
        args: argparse.Namespace,
        feed_list: List[Text],
        partial: bool) -> List[Text]:

    """Download and analyze a list of feeds concurrently"""

    files: List[Text] = []

    with concurrent.futures.ThreadPoolExecutor(max_workers=32) as executor:
        # Start the load operations and mark each future with its URL
        future_to_url = {executor.submit(handle_feed, args, url, partial): url
                         for url in feed_list}
        for future in concurrent.futures.as_completed(future_to_url):
            url = future_to_url[future]
            result, feed, files = future_result_exception(url, future)
            logging.info("Feed[%s] (%s) returned %s with %s files", feed, url, result, len(files))
            files += files

    return files


def future_result_exception(url: Text,
                            future: concurrent.futures.Future) -> Tuple[Text, Text, List[Text]]:
    """Retrieve the result of a future, logging any exception"""

    try:
        result, feed, files = future.result()
        logging.info("%s returned %s",
                     feed,
                     result)
        return result, feed, files
    except Exception as exc:  # pylint: disable=W0703
        logging.error('%r generated an exception: %s', url, exc)
        exc_info = (type(exc), exc, exc.__traceback__)
        logging.error('Exception occurred', exc_info=exc_info)
        return "EXCEPTION", "NA", []



"""Helper function for data extraction"""

from typing import Dict, Text, Tuple, Optional, List
import logging
import argparse
import os.path
import urllib.parse

from bs4 import BeautifulSoup
import html
import justext
import requests

from act.scio.feeds import download, analyze


def get_content_from_entry(entry: Dict) -> Text:
    """Extract and concatenate the content portion of a feed entry"""

    if "content" in entry:
        return "\n".join([x["value"] for x in entry['content']])

    logging.warning("No content for %s", entry['link'])

    return ""


def get_summary_from_entry(entry: Dict) -> Text:
    """Extract summary from feed entry"""

    if "summary_detail" in entry:
        value: Text = entry['summary_detail']['value']
        return value
    return ""


def create_html(entry: Dict) -> Text:
    """Wrap an entry in html headers and footers"""

    html_data = """<html>
    <head>
        <title>{title}</title>
    </head>
    <body>
        <p>
        {summary}
        </p>
        {content}
    </body>
</html>"""

    content = get_content_from_entry(entry)
    summary = get_summary_from_entry(entry)

    return html_data.format(title=entry["title"],
                            content=content,
                            summary=summary)


def safe_filename(path: Text) -> Text:
    """Make filename safe by only allowing alpha numeric characters,
    digits and ._-"""

    def _safe_char(c: Text) -> Text:
        if c.isalpha():
            return c
        if c.isdigit():
            return c
        if c in "_-.":
            return c
        if c in " \t":
            return "_"
        return ""

    return "".join(_safe_char(c) for c in path)


def read_stoplist(fname: Text) -> frozenset:
    """Take a list of words (one pr. line) and create a frozenset
    for use as stopwords when extracting text with jusText"""

    with open(fname, "r") as stream:
        return frozenset(line.strip().lower() for line in stream)


def partial_entry_text_to_file(
        args: argparse.Namespace,
        entry: Dict) -> Tuple[Optional[Text], Optional[Text]]:
    """Download the original content and write it to the proper file.
    Return the html."""

    if "link" not in entry:
        logging.warning("entry does not contain 'link'")
        return None, None

    url = entry["link"]

    req = requests.get(url, headers=download.default_headers(), verify=False, timeout=60)

    if req.status_code >= 400:
        logging.warning("Unable to download content: %s", url)
        return None, None

    filename = safe_filename(entry['title'])

    html_data = "<html>\n<head>\n"
    html_data += "<title>{0}</title>\n</head>\n".format(entry['title'])
    html_data += "<body>\n"

    raw_html = req.text

    stoplist = read_stoplist(args.stoplist) if args.stoplist else justext.get_stoplist('English')

    paragraphs = justext.justext(raw_html, stoplist)

    for para in paragraphs:
        if not para.is_boilerplate:
            if para.is_heading:
                html_data += "\n<h1>{0}</h1>\n".format(html.escape(para.text))
            else:
                html_data += "<p>\n{0}\n</p>\n".format(html.escape(para.text))

    html_data += "\n</body>\n</html>"

    full_filename = os.path.join(args.store_path, "download", filename + ".html")
    with open(full_filename, "w") as html_file:
        html_file.write(html_data)

    # we want to return the raw_html and not the "article extraction"
    # since we want to extract links to .pdfs etc.
    return full_filename, raw_html


def entry_text_to_file(
        args: argparse.Namespace,
        entry: Dict) -> Tuple[Optional[Text], Optional[Text]]:
    """Extract the entry content and write it to the proper file.
    Return the wrapped HTML"""

    filename = safe_filename(entry['title'])

    html_data = create_html(entry)

    full_filename = os.path.join(os.path.join(args.store_path, "download"), filename + ".html")
    with open(full_filename, "w") as html_file:
        html_file.write(html_data)

    return full_filename, html_data


def get_links(entry: Dict, html_data: Text) -> List[urllib.parse.ParseResult]:
    """Extract any links from the html"""

    links = []
    soup = BeautifulSoup(html_data, "html.parser")
    if soup:
        links = [a['href'] for a in soup.findAll('a', href=True)]
    else:
        logging.warning("soup is none : %s", entry['title'])

    return [analyze.parse_and_correct_link(link) for link in links]



import addict
from act.scio.aliasregex import normalize
from act.scio.vocabulary import Vocabulary
from act.scio.plugin import BasePlugin, Result
from typing import Text, List
import configparser
import os.path


def normalize_ta(name: Text) -> Text:
    return normalize(
        name,
        capitalize=True,
        uppercase_abbr=["APT", "BRONZE", "IRON", "GOLD"],
    )


class Plugin(BasePlugin):
    name = "threatactor"
    info = "Extracting references to known threat actors from a body of text"
    version = "0.2"
    dependencies: List[Text] = []

    async def analyze(self, nlpdata: addict.Dict) -> Result:

        ini = configparser.ConfigParser()
        ini.read([os.path.join(self.configdir, "threatactor_pattern.ini")])
        ini['threat_actor']['alias'] = os.path.join(self.configdir, ini['threat_actor']['alias'])

        vocab = Vocabulary(ini['threat_actor'])

        res = addict.Dict()

        res.ThreatActors = vocab.regex_search(
            nlpdata.content,
            normalize_result=normalize_ta,
            debug=self.debug)

        return Result(name=self.name, version=self.version, result=res)
